"""Utilities for PRCS administration."""
# based on cvslib.py
# The main difference bewteen of prcs and cvs on the client side is:
#     prcs has a project file (descriptor) which has all the administration 
#     infomation of a prcs project
#     while cvs stores them in CVS subdirectories.

import os
import re
import string
import pickle
from workdir import DirSupport, WorkDir


class PRCS(DirSupport):

    """
    PRCS interface class (local filesystem version)
    It will operate on a list of files

    It works in the current directory..
       
    variables:
    project:      project
    rev:          its revision (string)
                  if revision is None, the latest revision is used 
    files:        a list of filenames ([] if no file is specified)
    name_rev:     a mapped rcs file and the rcs revision,
                  both are non-blank string
       
    """
	
    def __init__(self):
        """Constructor."""
	DirSupport.__init__(self)
	self._cache = {}
	self._TMPDIR = '/tmp'
	self._SCRATCH = '/tmp/.scratch'
    	env = os.environ
	if env.has_key('PRCS_REPOSITORY'):
	    self._PRCSHOME = env['PRCS_REPOSITORY']
	else:
	    self._PRCSHOME = env['HOME'] + '/PRCS'

    def __del__(self):
        """Destructor."""
        DirSupport.__del__(self)


    # --- Informational methods about a single file/revision ---

    def list(self):
	"""
	return the available projects in the repository
	"""
	self.cd(self._PRCSHOME)
	files = self.listsubdirs(igns = ['.*'])
	self.back()
	return files
	
    def revtree(self, project):
	"""
	parse a rlog of the prjfile, and generate a dict of dict:
	branch(major version): dict of minor version -> infomation
	"""
	re_majorrev = re.compile(r"^PRCS major version: (\S+)")
	re_minorrev = re.compile(r"^PRCS minor version: (\S+)")
	re_parentid = re.compile(r"^PRCS parent indices: (\S+)")
	re_deleted = re.compile(r"^PRCS version deleted")
	dict = {}
	i = 0				# the smallest i, the latest
	
	cur_majotrev = 0
	stream = self._open((self._prjfile(project), None), 'rlog')
	while 1:
	    line = stream.readline()
	    if not line: break
	    while not re_majorrev.match(line):
		line = stream.readline()
	    m = re_majorrev.match(line)
	    line = stream.readline()
	    n = re_minorrev.match(line) # should always match
	    line = stream.readline()
	    p = re_parentid.match(line)
	    line = stream.readline()
	    q = re_deleted.match(line)
	    if q: o = None
	    else: o = 'Exist'
	    try:
		dict[m.group(1)][n.group(1)] = (o, i)
		i = i + 1
	    except KeyError:
		dict[m.group(1)] = {}
		dict[m.group(1)][n.group(1)] = (o, i)
		i = i + 1
        status = self._closepipe(stream)
        if status:
             raise IOError, status
	return dict

    def getlatestrevid(self, project):
	" Only exception here: rev can be None "
	dict = self.revtree(project)
	# seeking the entry with lowest time stamp
	i = 0
	while 1:
	    for Mrev in dict.keys():
		for mrev in dict[Mrev].keys():
		    if dict[Mrev][mrev][0] and dict[Mrev][mrev][1] == i:
			return Mrev + '.' + mrev
	    i = i + 1
	    
    def getbranchhead(self, project, rev):
	if not rev:
	    return self.getlatestrevid(project)
	
	p = string.find(rev, '.')
	if p < 0: majorrev = rev
	else: majorrev = rev[:p]
	dict = self.revtree(project)
	minorrev = max(map(string.atoi, dict[majorrev].keys()))
	while not dict[majorrev][`minorrev`][0]:
	    minorrev = minorrev - 1
        return majorrev + '.' + `minorrev`

    def normalizerev(self, project, rev):
	if rev == None:
	    return self.getlatestrevid(project)
	elif string.find(rev, '.') < 0:
	    return self.getbranchhead(project, rev)
	return rev
	
    def getunkeyedsums(self, project, rev=None):
	"""It is unkeyed md5sums"""
	rev = self.normalizerev(project, rev)
	dict = {}
	f = self._open2(project, rev, cmd = 'prcs info -L')
	re_filesum = re.compile(r"(\S+)\s+MD5=(\S+)")
	while 1:
	    line = f.readline()
            if not line: break
	    m = re_filesum.match(line)
	    if m:
                dict[m.group(1)] = m.group(2)
        status = self._closepipe(f)
        if status:
            raise IOError, status
        return dict

    def getsums(self, project, rev=None):
	"""It is keyed md5sum"""
	rev = self.normalizerev(project, rev)
	dict = {}
	files = self._descriptor(project, rev).keys()
	self.mkdir(self._SCRATCH)
	self.cd(self._SCRATCH)
	self.checkout(project, rev, files)
	for file in files:
	    dict[file] = self.sum(file)
	self.back()
	cmd = 'rm -rf %s' % self._SCRATCH
	self._system(cmd)
	return dict

    def info(self, project, rev=None, files=[], otherflags = ''):
        """Return the log for specified project version
        """
	f = self._open2(project, rev, files, 'prcs info ' + otherflags)
	data = f.read()
        status = self._closepipe(f)
        if status:
            data = data + "%s: %s" % status
        elif data[-1] == '\n':
            data = data[:-1]
        return data
    
    def log(self, project, rev=None, files=[], rflag = '-r'):
        """Return the full log text for NAME_REV as a string.

	NAME_REV is the mapped rcs file name and its revision
	
        Optional OTHERFLAGS are passed to rlog.

        """
	rev = self.normalizerev(project, rev)
	data = ''
	for file in files:
	    name_rev = self._namerev(project, rev, file)
	    data = data + self._log(name_rev, rflag)
	return data
    

    # --- Methods that change files ---

    def checkout(self, project, rev=None, files=[], otherflags = ''):
        """Check out projetc_rev_file to its work file.

        The optional OTHERFLAGS is passed to co without
        interpretation.

        Any output from co goes to directly to stdout.

        """
	filelist = reduce(lambda x, y: x+' '+y, files, '')
	if rev:
	    cmd = 'prcs checkout -fr%s %s %s %s' % (rev, otherflags, project, filelist)
	else:
	    cmd = 'prcs checkout -f %s %s %s' % (otherflags, project, filelist)
        return self._system(cmd)

    def checkin(self, project, rev=None, files=[], otherflags = ''):
        """Check in project from its work file.

        The optional MESSAGE argument becomes the checkin message
        (default "<none>" if None); or the file description if this is
        a new file.

        The optional OTHERFLAGS argument is passed to ci without
        interpretation.

        Any output from ci goes to directly to stdout.

        """
	filelist = reduce(lambda x, y: x+' '+y, files, '')
	if rev:
	    cmd = 'prcs checkin -fr %s %s %s %s' % (rev, otherflags, project, filelist)
	else:
	    cmd = 'prcs checkin -f %s %s %s' % (otherflags, project, filelist)
	self.cd(self._TMPDIR)
	msg = self._system(cmd)
	self.back()
	return msg

    def edit(self, project, rev, file, log = None):
	"""Use cmd rcs -mrev:msg to replace a file's log with log
	"""
	rev = self.normalizerev(project, rev)
	name, rev = self._namerev(project, rev, file)
	print name
	if log:
	    cmd = 'rcs -m%s:\"%s\" %s' % (rev, log, name) 
	    print cmd
	return self._system(cmd)

    # --- Internal methods ---

    def _prjfile(self, project):
	return self._PRCSHOME + '/' + project + '/' + project + '.prj,v'

    def _descriptor(self, project, rev):
	"""Return the descriptor """
	rev = self.normalizerev(project, rev)
	try:
	    return self._cache[(project, rev)]
	except KeyError:
	    self.cd(self._TMPDIR)
	    prjfile = project + '.prj'
	    cmd = 'prcs checkout -fr%s %s %s' % (rev, project, prjfile)
	    self._system(cmd)
	    f = open(prjfile)
	    self._cache[(project, rev)] = descriptor(f)[3]
	    f.close()
	    self._remove(prjfile)
	    self.back()
	    return self._cache[(project, rev)]


    def _namerev(self, project, rev, file):
	"""Return the rcs mapped name and its revision"""
	try:
	    name, rev = self._descriptor(project, rev)[file]
	    return (self._PRCSHOME + '/' + name + ',v', rev)
	except KeyError:
	    raise IOError, "%s is not found in the specified project %s version %s." % (file, project, rev)

    def _info(self, name_rev):
        """Return a dictionary of info (from rlog -h) for NAME_REV

        The dictionary's keys are the keywords that rlog prints
        (e.g. 'head' and its values are the corresponding data
        (e.g. '1.3').

        XXX symbolic names and locks are not returned

        """
        f = self._open(name_rev, 'rlog -h')
        dict = {}
        while 1:
            line = f.readline()
            if not line: break
            if line[0] == '\t':
                # XXX could be a lock or symbolic name
                # Anything else?
                continue 
            i = string.find(line, ':')
            if i > 0:
                key, value = line[:i], string.strip(line[i+1:])
		dict[key] = value
	status = self._closepipe(f)
        if status:
            raise IOError, status
        return dict

    def _log(self, name_rev, rflag = '-r'):
        """Return the full log text for NAME_REV as a string.

        Optional OTHERFLAGS are passed to rlog.

        """
        f = self._open(name_rev, 'rlog', rflag)
        data = f.read()
        status = self._closepipe(f)
        if status:
            data = data + "%s: %s" % status
        elif data[-1] == '\n':
            data = data[:-1]
        return data
 

    def _open(self, name_rev, cmd = 'co -p', rflag = '-r'):
        """INTERNAL: open a read pipe to NAME_REV using optional COMMAND.

        Optional FLAG is used to indicate other command flags (default -r).

        Default COMMAND is "co -p".

        Return a file object connected by a pipe to the command's
        output.

        """
        name, rev = name_rev
	if rev:
	    cmd = cmd + ' ' + rflag + rev
        return os.popen("%s %s" % (cmd, name))

    def _open2(self, project, rev, files = [], cmd = 'prcs info', rflag = '-r'):
        """INTERNAL: open a read pipe to Project using optional COMMAND.

        Optional FLAG is used to indicate other command flags (default -r)

        Default COMMAND is "prcs info".

        Return a file object connected by a pipe to the command's
        output.

        """
	filelist = reduce(lambda x, y: x+' '+y, files, '')
	if rev:
	    cmd = cmd + ' ' + rflag + rev
        return os.popen("%s %s %s" % (cmd, project, filelist))

    def _closepipe(self, f):
        """INTERNAL: Close PIPE and print its exit status if nonzero."""
        sts = f.close()
        if not sts: return None
        detail, reason = divmod(sts, 256)
        if reason == 0: return 'exit', detail   # Exit status
        signal = reason&0x7F
        if signal == 0x7F:
            code = 'stopped'
            signal = detail
        else:
            code = 'killed'
        if reason&0x80:
            code = code + '(coredump)'
        return code, signal
 
    def _system(self, cmd):
        """INTERNAL: run COMMAND in a subshell.

        Standard input for the command is taken fron /dev/null.

        Raise IOError when the exit status is not zero.

        Return whatever the calling method should return; normally
        None.

        A derived class may override this method and redefine it to
        capture stdout/stderr of the command and return it.

        """
        cmd = cmd + " </dev/null"
        sts = os.system(cmd)
        if sts: raise IOError, "command exit status %d" % sts

    def _remove(self, fn):
        """INTERNAL: remove FILE without complaints."""
        try:
            os.unlink(fn)
        except os.error:
            pass




class PRCSWorkDir(WorkDir):

    """
    PRCS WorkDir is a working directory plus a controlling file
    project.prj, which must present.
    """

    def __init__(self):
	WorkDir.__init__(self)
	self.proxy = None
	self.project = self.getdefaultproject()
	self.version = self.getdefaultversion()

    def setproxy(self, proxy):
	self.proxy = proxy
		      
    def getdefaultproject(self):
	"Return the default project in current directory"
	prjfiles = self.listfiles(pats = ["*.prj"])
	if len(prjfiles) == 1:
	    return prjfiles[0][:-4]
	else:
	    raise IOError, 'None or multiple prjfiles present'

    def getdefaultversion(self):
	return descriptor(open(self.project + '.prj'))[1]

    def getdescriptor(self):
	return descriptor(open(self.project + '.prj'))[3]

    def filesinproject(self):
	return self.getdescriptor().keys()
    
    def status(self):
	""" the report should combine the local descriptor and WorkDir.status
	    It return (an expanded descriptor with md5sum,
	               and a list of extra files)
            md5sum = None means that file is missing
	"""
	descriptor = self.getdescriptor()
	wd = WorkDir.status(self)
	dict = {}
	extra = wd.keys()
	for f in descriptor.keys():
	    if f in wd.keys():
		dict[f] = (descriptor[f][0], descriptor[f][1], wd[f])
		extra.remove(f)
	    else:
		dict[f] = (descriptor[f][0], descriptor[f][1], None)
	return (dict, extra)
		     
    def getrsums(self, rev):
	if not rev:
	    rev = self.proxy.getlatestrevid(self.project)
	if string.find(rev, '.') < 0:
	    rev = self.proxy.getbranchhead(self.project, rev)
	rstatefile = '.' + self.project + '_' + rev + '_state'
	if os.path.exists(rstatefile):
	    return pickle.load(open(rstatefile))
	else:
	    dict = self.proxy.getsums(self.project, rev)
	    pickle.dump(dict, open(rstatefile, 'w'))
	    return dict

    def check(self, rev, files = []):

	"""Report on a file's status:
	
	if files are present, the report on these files are returned, else
	a report on all available files are returned.
	
	two levels of check:
	consistancy:      Missing and New
	modified status:  Modified, Uptodate, Added and Removed
	
	The 1st level checks the local files physically present with the prjfile
	The 2nd level checks the local and remote prjfiles
	
	It takes a local view.
	
	Status:
	'???' -- Unknown, not in working direction or local project file
	         or remote project file

	'Omt' -- Missing
	'Xtr' -- new local file
	
	'Mod' -- Changed locally
	'===' -- Uptodate, everything's fine
	'Add' -- Added locally
	'Rmv' -- Removed locally
	
	'Prj' -- Prjfile

	"""
	
	local, lextra = self.status()
	remote = self.getrsums(rev)
	lfiles = local.keys()
	rfiles = remote.keys()
	dict = {}


	if files:
	    for f in files:
		if f == self.project + '.prj':
		    dict[f] = 'Prj'
		elif f in lfiles and f in rfiles:
		    if local[f][2] == remote[f]:
			dict[f] = '==='
		    else:
			dict[f] = local[f][2] == None and 'Omt' or 'Mod'
		elif f in rfiles:
		    dict[f] = 'Rmv'
		elif f in lfiles:
		    dict[f] = local[f][2] == None and 'Omt' or 'Add'
		elif f in lextra:
		    dict[f] = 'Xtr'
		else:
		    dict[f] = '???'
	    return dict

	else:
	    for f in lfiles:
		if f in rfiles:
		    if remote[f] == local[f][2]:
			dict[f] = '==='
		    else:
			dict[f] = local[f][2] == None and 'Omt' or 'Mod'
		    rfiles.remove(f)
		else:
		    dict[f] = 'Add'
	    for f in rfiles:
		dict[f] = 'Rmv'
	    for f in lextra:
		dict[f] = 'Xtr'
	    return dict



# it is not in the PRCS class, since it will also be used to parse
# woking prjfile.

def descriptor(stream):
    
    """
    Descriptor Object, maps all the information in a descriptor file
    (.prj file) into attributes of the object descriptor
    
    attributes:

    stream: 	  an opened readonly file object, locally opened file
                  or _socket.makefile('r')
    description:  description of the project
    project:      the name of the project
    version:      the version
    dict:         a dictionary stores all the files in the project with
                  /file: mapped_file  its mapped file in repository,\ 
                  \      and its corresponding rcs version          /
    """
    
    dict = {}
    re_description_begin = re.compile(r"^\(Project-Description")
    re_description_end = re.compile(r".*\"\)$", re.M + re.S)
    re_description = re.compile(r"^\(Project-Description \"(.*)\"\)$", re.M + re.S)
    re_projectversion = re.compile(r"^\(Project-Version (\S+) (\S+) ([0-9.]+)")
	
    re_file = re.compile(r"^\s*\((\S+)\s+\((\S+)\s+(\S+)\s\S+\)")
    re_freshfile = re.compile(r"^\s+\((\S+)\s+\(\)\)$")
    while 1:
	line = stream.readline()
	if not line: break
    			
	if re_description_begin.match(line):
	    while not re_description_end.match(line):
		line = line + stream.readline()
	    description = re_description.match(line).group(1)
	elif re_projectversion.match(line):
	    m = re_projectversion.match(line)
	    project = m.group(1)
	    version = m.group(2)+'.'+m.group(3)
	elif re_file.match(line):
	    m = re_file.match(line)
	    dict[m.group(1)] = (m.group(2), m.group(3))
	elif re_freshfile.match(line):
	    m = re_freshfile.match(line)
	    dict[m.group(1)] = (None, None)
    return (project, version, description, dict)
